import numpy as np
from matplotlib import pyplot as plt
import os
from sklearn.inspection import DecisionBoundaryDisplay
from graphviz import Source
from sklearn.tree import export_graphviz

def plot2D_knn_results(knn_clf, X, y, feature_names, target_names, res=100, test_point=None):
    """  
    Muestra las regiones de decisión de un clasificador kNN para sus datos de entrenamiento (X,y).
    Al agregar un dato de prueba (con las mismas dimensiones de features que el conjunto de entrenamiento),
    se muestran sus vecinos más cercanos.
    """

    # Pintar regiones de decisión
    disp = DecisionBoundaryDisplay.from_estimator(
        knn_clf,
        X,        
        xlabel=feature_names[0],
        ylabel=feature_names[1],
        grid_resolution=res,
        response_method="predict",
        plot_method="pcolormesh",
        shading="auto",
        alpha=0.5
    )
    # Graficar muestras
    scatter = disp.ax_.scatter(X[:, 0], X[:, 1], c=y, edgecolors="k")
    _ = disp.ax_.set_title(
        f"Clasificación\n(k={knn_clf.n_neighbors} vecinos)"
    )
    first_legend = disp.ax_.legend(
        scatter.legend_elements()[0],
        target_names,
        loc="best",
        title="Clases",
    )

    disp.ax_.add_artist(first_legend)

    # Mostrar dato de prueba y sus vecinos más cercanos
    if test_point is not None:
        # Mostrar dato de prueba
        test_dot, = plt.plot(test_point[0],test_point[1],'or', label='Dato test')
        # Obtener y mostrar sus vecinos más cercanos
        nearest_kneighbors = X[knn_clf.kneighbors( [test_point], knn_clf.n_neighbors,return_distance=False)][0]
        nearest_kneighbors_labels = y[knn_clf.kneighbors( [test_point], knn_clf.n_neighbors,return_distance=False)]
        scatter_knn = disp.ax_.scatter(nearest_kneighbors[:, 0], nearest_kneighbors[:, 1], s=100, color=(0,0,0,0), edgecolors="tab:orange", label='Vecinos más cercanos')
        knn_legend = disp.ax_.legend(
            scatter_knn.legend_elements()[0],
            target_names,
            loc='lower right',
            title="Vecinos más cercanos",
        )
        last_legend = disp.ax_.legend(handles=[test_dot,scatter_knn],loc='lower right')    
        disp.ax_.add_artist(last_legend)
    
    plt.show()

# Función para graficar desplegar forma de árbol de decisión
def display_tree(tree_clf, img_path, filename):
    export_graphviz(
            tree_clf,
            out_file=os.path.join(img_path, filename),
            rounded=True,
            filled=True
        )

    display(Source.from_file(os.path.join(img_path, filename)))

# Función para entrenar y graficar regiones de decisión
def plot_classifier_regions(clf, 
                      X_train, y_train,
                      X_test, y_test
                      ):
    # Ajustar árbol
    clf.fit(X_train, y_train)
    score_train = clf.score(X_train, y_train)
    score_test = clf.score(X_test, y_test)
    
    X = np.vstack([X_train, X_test])
    y = np.hstack([y_train, y_test])

    # Crea una malla de puntos para visualizar las regiones de decisión
    x_min, x_max = np.min(X) - 1, np.max(X) + 1
    y_min, y_max = np.min(X) - 1, np.max(X) + 1
    xx, yy = np.meshgrid(np.linspace(x_min, x_max, 500), np.linspace(y_min, y_max, 500))
    Z = clf.predict(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)

    # Grafica las regiones de decisión
    plt.figure(figsize=(5, 3))
    plt.contourf(xx, yy, Z, alpha=0.4)
    plt.scatter(X[:, 0], X[:, 1], c=y, s=20, edgecolor='k')
    plt.xlabel('Feature 1')
    plt.ylabel('Feature 2')
    plt.title(f'Decision Boundary of Classifier (Train = {score_train:.2f}, Test = {score_test:.2f})')
    plt.show()